import pickle
from ml_instrumentation.Collector import Collector
from ml_instrumentation.Writer import SqlPoint

def test_collector_rw1(basic_collector):
    basic_collector.set_experiment_id(0)
    basic_collector.next_frame()

    basic_collector.collect('m1', 0)
    assert basic_collector.get('m1', 0) == [
        SqlPoint(frame=0, id=0, measurement=0)
    ]

    basic_collector.next_frame()
    basic_collector.collect('m1', 2)
    basic_collector.collect('m2', 1)
    assert basic_collector.get('m1', 0) == [
        SqlPoint(frame=0, id=0, measurement=0),
        SqlPoint(frame=1, id=0, measurement=2),
    ]

    assert basic_collector.get('m2', 0) == [
        SqlPoint(frame=1, id=0, measurement=1),
    ]

def test_collector_rw_trailing_edge1(basic_collector):
    basic_collector.set_experiment_id(0)

    # note: no next_frame call here, it will go after
    basic_collector.collect('m1', 0)
    assert basic_collector.get('m1', 0) == [
        SqlPoint(frame=0, id=0, measurement=0),
    ]

    basic_collector.next_frame()

    # next_frame can also go in between collect and get
    basic_collector.collect('m1', 1)
    basic_collector.next_frame()
    assert basic_collector.get('m1', 0) == [
        SqlPoint(frame=0, id=0, measurement=0),
        SqlPoint(frame=1, id=0, measurement=1),
    ]


def test_collector_serde(basic_collector):
    basic_collector.set_experiment_id(0)
    basic_collector.next_frame()

    basic_collector.collect('m1', 0)
    basic_collector.next_frame()
    basic_collector.collect('m1', 1)

    byts = pickle.dumps(basic_collector)
    collector2: Collector = pickle.loads(byts)

    data1 = basic_collector.get('m1', 0)
    data2 = collector2.get('m1', 0)
    assert data1 == data2 == [
        SqlPoint(frame=0, id=0, measurement=0),
        SqlPoint(frame=1, id=0, measurement=1),
    ]

    collector2.close()



import sqlite3
from pathlib import Path
from typing import Any

from filelock import FileLock

import ml_instrumentation._utils.sqlite as sqlu

def attach_metadata(db_path: str | Path, id: int | str, metadata: dict[str, Any]):
    db_path = Path(db_path)
    db_path.parent.mkdir(parents=True, exist_ok=True)

    with FileLock(f'{db_path}.lock'):
        con = sqlite3.connect(db_path)
        cur = con.cursor()

        metadata |= {'id': id}
        cols = metadata.keys()
        sqlu.maybe_make_table(cur, '_metadata_', cols)
        sqlu.ensure_table_compatible(cur, '_metadata_', cols)

        inserter = ', '.join(['?'] * len(cols))

        header = ', '.join(map(sqlu.quote, cols))
        query = f'INSERT INTO _metadata_ ({header}) VALUES ({inserter})'
        print(query)
        cur.execute(query, list(metadata.values()))
        con.commit()
        con.close()



from functools import partial
import jax
import jax.experimental
from ml_instrumentation.Collector import Collector
from ml_instrumentation.Writer import SqlPoint

def test_collector_jax_jit(basic_collector: Collector):
    basic_collector.set_experiment_id(0)
    basic_collector.next_frame()

    @partial(jax.jit, static_argnums=(1,))
    def f(x: jax.Array, collector: Collector):
        y = 2 * x
        y = y.dot(y)
        collector.collect_jax('m1', y)
        return y

    x = jax.numpy.array([0, 1, 2, 3])
    y = f(x, basic_collector)

    assert y == 56
    assert basic_collector.get('m1', 0) == [
        SqlPoint(frame=0, id=0, measurement=56)
    ]

def test_collector_jax_grad(basic_collector: Collector):
    basic_collector.set_experiment_id(0)
    basic_collector.next_frame()

    @partial(jax.jit, static_argnums=(1,))
    @jax.grad
    def f(x: jax.Array, collector: Collector):
        y = 2 * x
        y = y.dot(y)
        collector.collect_jax('m1', y)
        return y

    x = jax.numpy.array([0, 1, 2, 3.])
    f(x, basic_collector)

    assert basic_collector.get('m1', 0) == [
        SqlPoint(frame=0, id=0, measurement=56)
    ]

def test_collector_jax_vmap(basic_collector: Collector):
    basic_collector.set_experiment_id(0)
    basic_collector.next_frame()

    @partial(jax.jit, static_argnums=(1,))
    @partial(jax.vmap, in_axes=(0, None), out_axes=0)
    def f(x: jax.Array, collector: Collector):
        y = 2 * x
        y = y.dot(y)
        collector.collect_jax('m1', y)
        return y

    x = jax.numpy.array([
        [0, 1, 2, 3.],
        [1, 1, 2, 3.],
    ])
    f(x, basic_collector)

    assert basic_collector.get('m1', 0) == [
        SqlPoint(frame=0, id=0, measurement=56),
        SqlPoint(frame=0, id=0, measurement=60),
    ]



from collections.abc import Iterable
from typing import Any
from sqlite3 import Cursor


def get_tables(cur: Cursor) -> set[str]:
    cur.row_factory = None
    res = cur.execute("SELECT name FROM sqlite_master")
    return set(r[0] for r in res.fetchall())


def make_table(cur: Cursor, name: str, columns: Iterable[str]):
    cols = ', '.join(map(quote, columns))
    cur.execute(f'CREATE TABLE {name}({cols})')


def maybe_make_table(cur: Cursor, name: str, columns: Iterable[str]):
    tables = get_tables(cur)

    if name not in tables:
        make_table(cur, name, columns)


def get_cols(cur: Cursor, name: str):
    res = cur.execute(f'PRAGMA table_info({name})')
    rows = res.fetchall()

    return set(r[1] for r in rows)


def add_cols(cur: Cursor, columns: Iterable[str]):
    columns = map(quote, columns)
    for col in columns:
        cur.execute(f'ALTER TABLE results ADD COLUMN {col}')


def ensure_table_compatible(cur: Cursor, name: str, columns: Iterable[str]):
    columns = set(columns)
    current_cols = get_cols(cur, name)
    needed_cols = columns - current_cols

    if needed_cols:
        add_cols(cur, needed_cols)


def maybe_quote(v: Any):
    if isinstance(v, str):
        return quote(v)
    return v

def quote(s: str):
    return f'"{s}"'



from functools import partial
import sqlite3
from ml_instrumentation.Writer import Point, SqlPoint, Writer
from multiprocessing.pool import Pool

def test_write1(writer):
    d = Point(
        metric='measurement-1',
        exp_id=0,
        frame=0,
        data=1.1,
    )
    writer.write(d)

    # should get back a list of datapoints for this measurement
    points = writer.read_metric('measurement-1')
    assert points == [
        SqlPoint(frame=0, id=0, measurement=1.1),
    ]

    d = Point(
        metric='measurement-1',
        exp_id=0,
        frame=1,
        data=2.2
    )
    writer.write(d)

    points = writer.read_metric('measurement-1')
    assert points == [
        SqlPoint(0, 0, 1.1),
        SqlPoint(1, 0, 2.2),
    ]

def test_write2(writer):
    for i in range(1_000):
        d = Point(
            metric='measurement-1',
            exp_id=0,
            frame=i,
            data=i * 1.1,
        )
        writer.write(d)

        if i % 3 == 0:
            d2 = d._replace(metric='measurement-2')
            writer.write(d2)

    points = writer.read_metric('measurement-1')
    assert len(points) == 1_000
    for i in range(1_000):
        assert points[i].frame == i
        assert points[i].measurement == i * 1.1

    points = writer.read_metric('measurement-2')
    assert len(points) == 334


def test_merge1(tmp_path):
    w1 = Writer(
        db_path=str(tmp_path / 'w1.db'),
    )

    w2 = Writer(
        db_path=':memory:',
    )

    for i in range(10):
        w1.write(Point(exp_id=0, metric='a', frame=i, data=i))
        w2.write(Point(exp_id=1, metric='a', frame=i, data=100+i))

    w1.merge(str(tmp_path / 'total.db'))
    w2.merge(str(tmp_path / 'total.db'))

    con = sqlite3.connect(tmp_path / 'total.db')
    cur = con.cursor()

    cur.execute('SELECT * FROM a WHERE id=0 ORDER BY frame')
    res = cur.fetchall()
    assert len(res) == 10
    assert [r[2] for r in res] == [i for i in range(10)]

    cur.execute('SELECT * FROM a WHERE id=1 ORDER BY frame')
    res = cur.fetchall()
    assert len(res) == 10
    assert [r[2] for r in res] == [100+i for i in range(10)]

    cur.execute('SELECT * FROM a')
    res = cur.fetchall()
    assert len(res) == 20

    w1.close()
    w2.close()


def test_merge_parallel1(tmp_path):
    pool = Pool(10)
    pool.map(partial(_test_merge_parallel1, tmp_path=tmp_path), range(20))

    con = sqlite3.connect(tmp_path / 'total.db')
    cur = con.cursor()

    for i in range(20):
        cur.execute(f'SELECT * FROM a WHERE id={i} ORDER BY frame')
        res = cur.fetchall()
        assert len(res) == 100
        assert [r[2] for r in res] == [2*j + i for j in range(100)]

        cur.execute(f'SELECT * FROM b where id={i} ORDER BY frame')
        res = cur.fetchall()
        assert len(res) == 100
        assert [r[2] for r in res] == [f'{j} - {i}' for j in range(100)]

    cur.execute('SELECT * FROM a')
    res = cur.fetchall()
    assert len(res) == 20 * 100


def _test_merge_parallel1(i: int, tmp_path):
    writer = Writer(db_path=str(tmp_path / f'w{i}.db'))

    for j in range(100):
        writer.write(Point(exp_id=i, metric='a', frame=j, data=2*j + i))
        writer.write(Point(exp_id=i, metric='b', frame=j, data=f'{j} - {i}'))

    writer.merge(str(tmp_path / 'total.db'))



import pickle
import pytest

from ml_instrumentation.Collector import Collector
from tests.fixtures.collector import basic_collector, disk_collector, simulate_run, SimMetric

"""
User should be able to:
1. Build a collector object
2. Add data into the collector (possibly sparsely)
3. Interrupt the program flow and serialize the collector to disk
4. Load the collector from disk
5. Continue execution of the code and data collection
6. Load all results of collection for later manipulation

The above should be able to run with in-memory collection and disk-backed collection.
"""
@pytest.mark.parametrize('collector_fixture', [basic_collector, disk_collector])
def test_collector1(collector_fixture, tmp_path, request):
    collector: Collector = request.getfixturevalue(collector_fixture.__name__)
    collector.set_experiment_id(0)

    metrics = [
        SimMetric(
            metric='m1',
            pred=lambda i: i % 100 == 0,
            value=lambda i: i,
        ),
        SimMetric(
            metric='m2',
            pred=lambda i: i % 33 == 0,
            value=lambda i: f'test string {i}',
        )
    ]

    # execute the first half of the program
    expected_1 = simulate_run(
        collector,
        start=0, end=50_000,
        metrics=metrics,
    )

    # serialize/deserialize
    with open(tmp_path / 'chk.pkl', 'wb') as f:
        pickle.dump(collector, f)

    del collector

    with open(tmp_path / 'chk.pkl', 'rb') as f:
        collector = pickle.load(f)

    assert collector.get_frame() == (50_000 - 1)

    # execute the second half of the program
    expected_2 = simulate_run(
        collector,
        start=50_000, end=100_000,
        metrics=metrics,
    )

    # load all data for evaluation
    a = collector.get('m1', experiment_id=0)
    assert len(a) == 1_000

    expected_a = expected_1['m1'] + expected_2['m1']
    assert a == expected_a

    b = collector.get('m2', experiment_id=0)
    assert len(b) == 3_031

    expected_b = expected_1['m2'] + expected_2['m2']
    assert b == expected_b



from collections.abc import Iterable
from functools import reduce
from pathlib import Path
from typing import Any
import connectorx as cx
import sqlite3

import ml_instrumentation._utils.sqlite as sqlu


def read_to_df(db_path: str | Path, metric: str, ids: Iterable[int] | None = None):
    db_path = Path(db_path)

    constraints = ''
    if ids is not None:
        str_ids = map(str, map(sqlu.maybe_quote, ids))
        constraints = f'WHERE id IN ({",".join(str_ids)})'

    query = f'SELECT * FROM {metric} {constraints}'
    df = cx.read_sql(f'sqlite://{db_path}', query, partition_on='id', partition_num=4, return_type='polars')
    df = df.rename({'measurement': metric})
    return df


def read_metrics(db_path: str | Path, metrics: Iterable[str], ids: Iterable[int] | None = None):
    dfs = [read_to_df(db_path, metric, ids) for metric in metrics]
    df = reduce(lambda df1, df2: df1.join(df2, how='full', on=['id', 'frame']), dfs)
    return df


def load_all_results(db_path: str | Path, metrics: Iterable[str] | None = None, ids: Iterable[int] | None = None):
    con = sqlite3.connect(db_path)
    cur = con.cursor()

    tables = sqlu.get_tables(cur)
    if metrics is None:
        metrics = tables - {'_metadata_'}

    df = read_metrics(db_path, metrics, ids)

    if '_metadata_' not in tables:
        return df

    meta = cx.read_sql(f'sqlite://{db_path}', 'SELECT * FROM _metadata_', return_type='polars')
    return df.join(meta, how='left', on=['id'])


def get_run_ids(db_path: str | Path, params: dict[str, Any]):
    constraints = ' AND '.join(
        f'[{k}]={sqlu.maybe_quote(v)}' for k, v in params.items()
    )

    query = f'SELECT id FROM _metadata_ WHERE {constraints}'
    return cx.read_sql(f'sqlite://{db_path}', query, return_type='polars')['id'].to_list()
